parser_name_defaults = {
	'action_vt': '(Unnamed action)',
	'and_vt': '(Unnamed and)',
	'attr_bool_vt': '(Unnamed attr_bool)',
	'bind_vt': '(Unnamed bind)',
	'bits_vt': '(Unnamed bits)',
	'butnot_vt': '(Unnamed butnot)',
	'ch_vt': '(Unnamed ch)',
	'charset_vt': '(Unnamed charset)',
	'choice_vt': '(Unnamed choice)',
	'difference_vt': '(Unnamed difference)',
	'end_vt': '(Unnamed end)',
	'endianness_vt': '(Unnamed endianness)',
	'epsilon_vt': '(Unnamed epsilon)',
	'ignore_vt': '(Unnamed ignore)',
	'ignoreseq_vt': '(Unnamed ignoreseq)',
	'indirect_vt': '(Unnamed indirect)',
	'int_range_vt': '(Unnamed int_range)',
	'length_value_vt': '(Unnamed length_value)',
	'many_vt': '(Unnamed many)',
	'not_vt': '(Unnamed not)',
	'nothing_vt': '(Unnamed nothing)',
	'optional_vt': '(Unnamed optional)',
	'permutation_vt': '(Unnamed permutation)',
	'seek_vt': '(Unnamed seek)',
	'skip_vt': '(Unnamed skip)',
	'tell_vt': '(Unnamed tell)',
	'sequence_vt': '(Unnamed sequence)',
	'token_vt': '(Unnamed token)',
	'unimplemented_vt': '(Unnamed unimplemented)',
	'get_vt': '(Unnamed get)',
	'put_vt': '(Unnamed put)',
	'whitespace_vt': '(Unnamed whitespace)',
	'xor_vt': '(Unnamed xor)'
}

class VTTypes:
	# Intended to be initialized in InitParserBreakpoint
	def __init__(self):
		self.vt_symbols = {int(gdb.lookup_symbol(key)[0].value().address) : gdb.lookup_symbol(key)[0] for key in parser_name_defaults.keys()}

	# vt_p == vt_symbols[int(parser_vtable_p)].value().address
	# name = parser_name_defaults[vt_types.lookup_by_address(parser_vtable_p)].name
	# address is expected to be pointer to a parser that can be converted to int()
	# e.g. parser.address of a Parser object, integer
	def lookup_by_address(self, address):
		try:
			return self.vt_symbols[int(address)]
		except KeyError:
			return None

#vt_types = VTTypes()

# TODO: maybe save vtable type here on init, or potentially in Parser

# At this point in time, the parser Env classes are created in perform_lowlevel_parse. Ideally they could be initialized anywhere a HParser local named "parser" is accessible.
class HParserEnv:
	def __init__(self, parser, top_level_parse):
		self.top_level_parse = top_level_parse

	# parser is expected to be a Parser object
	def name_from_vtable(self, parser):
		parser_addr = parser.address
		# TODO: do this without passing a string to gdb.parse_and_eval()
		# perhaps using gdb.Value would be the best

		# turns out subscripting is allowed if the pointer is de referenced first, so this should work:
		# p_object = gdb.parse_and_eval("*parser")
		# p_env = p_object['env']

		# or:
		# p_env = gdb.parse_and_eval("parser").dereference()['env']
		hparser_t = gdb.lookup_type("HParser")
		hparser_p_t = hparser_t.pointer()

		# Construct a gdb.Value around the address we were given, and cast it to a HParser*
		parser_addr_val = gdb.Value(parser_addr)
		parser_p = parser_addr_val.cast(hparser_p_t)

		vtable_p = parser_p.dereference()['vtable']
		try:
			name = parser_name_defaults[self.top_level_parse.vt_types.lookup_by_address(vtable_p).name]
		# if lookup_by_address() returns None
		except AttributeError:
			name = "(Unknown parser type (vtable symbol not found in lookup)"
			print("Lookup fail for: " + str(hex(parser_addr))) # DEBUG
			print("vt_types: " + str(self.top_level_parse.vt_types)) # DEBUG
		except KeyError:
			name = "(Unknown parser type (vtable exists but has no default name associated)"

		return name

# TODO: Unit test:
# make SequenceEnv(parser, top_level_parse)
# parser_array = [top_level_parse.add_or_get_parser(gdb.parse_and_eval("((HSequence*) parser->env)->p_array[" + str(index) +"]")) for index in range(0, num_parsers)]
# compare pointers for equality

# TODO: consistent naming for member_parser, member_parser_pointers,

class AttrBoolEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_parser_p = gdb.parse_and_eval("((HAttrBool *) parser->env)->p")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj
		self.predicate_p = gdb.parse_and_eval("((HAttrBool *) parser->env)->pred")
		self.user_data_p = gdb.parse_and_eval("((HAttrBool *) parser->env)->user_data")

	def __str__(self):
		return str(self.member_parser)

class BindEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_parser_p = gdb.parse_and_eval("((BindEnv *) parser->env)->p")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj
		continuation_p = gdb.parse_and_eval("((BindEnv *) parser->env)->k")
		self.continuation = int(continuation_p)
		env_p = gdb.parse_and_eval("((BindEnv *) parser->env)->env")
		self.env = int(env_p)

	def __str__(self):
		return str([str(self.member_parser), str(hex(self.continuation)), str(hex(self.env))])

class BitsEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		self.length =  gdb.parse_and_eval("((struct bits_env *) parser->env)->length")
		self.signedp =  gdb.parse_and_eval("((struct bits_env *) parser->env)->signedp")
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return str([str(self.length), str(self.signedp)])

class ButNotEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		p1_p = gdb.parse_and_eval("((HTwoParsers *) parser->env)->p1")
		p1_obj = top_level_parse.add_or_get_parser(p1_p)
		if p1_obj.name is None:
			p1_obj.name_parser( self.name_from_vtable(p1_obj) )
		self.p1 = p1_obj
		p2_p = gdb.parse_and_eval("((HTwoParsers *) parser->env)->p2")
		p2_obj = top_level_parse.add_or_get_parser(p2_p)
		if p2_obj.name is None:
			p2_obj.name_parser( self.name_from_vtable(p2_obj) )
		self.p2 = p2_obj

	def __str__(self):
		return str([str(self.p1), str(self.p2)])

class ActionEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_parser_p = gdb.parse_and_eval("((HParseAction*) parser->env)->p")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj

		# should return a HAction
		action = gdb.parse_and_eval("((HParseAction*) parser->env)->action")
		self.action = int(action)
		user_data_p = gdb.parse_and_eval("((HParseAction*) parser->env)->user_data")
		self.user_data_p = int(user_data_p)

	def __str__(self):
		return str([str(self.member_parser), str(hex(self.action)), str(hex(self.user_data_p))])

class AndEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		member_parser_p = gdb.parse_and_eval("(HParser*) parser->env")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return str([str(self.member_parser)])

class ChEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		self.ch_arg = gdb.parse_and_eval("(uint8_t) parser->env")
		self.ch_value = int(self.ch_arg)
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return str(self.ch_value)

class CharsetEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		self.charset_arg = gdb.parse_and_eval("(HCharset) parser->env")
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return str(self.charset_arg)

# TODO: numeric parameter passed to parse_and_eval
class ChoiceEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		super().__init__(parser, top_level_parse)
		self.parser = parser
		self.member_parsers = []

		#frame = gdb.selected_frame()
		h_sequence_t = gdb.lookup_type("HSequence")
		h_sequence_p_t = h_sequence_t.pointer()

		# Get the parser object
		parser_obj = gdb.parse_and_eval("*parser")
		# Get parser_obj.env, cast it to HSequence *, then dereference
		p_env = parser_obj['env'].cast(h_sequence_p_t).dereference()
		# Get length of HSequence
		num_parsers = p_env['len']
		# Get the array of parsers in the sequence
		p_array = p_env['p_array']

		for index in range(0, num_parsers):
			parser_p = p_array[index]
			parser_obj = top_level_parse.add_or_get_parser(parser_p)
			if parser_obj.name is None:
				parser_obj.name_parser( self.name_from_vtable(parser_obj) )
			self.member_parsers.append(parser_obj)

	def __str__(self):
		return str([str(parser) for parser in self.member_parsers])

class DifferenceEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		#TODO: maybe have a class for HTwoParsers-based envs
		p1_p = gdb.parse_and_eval("((HTwoParsers *) parser->env)->p1")
		p1_obj = top_level_parse.add_or_get_parser(p1_p)
		if p1_obj.name is None:
			p1_obj.name_parser( self.name_from_vtable(p1_obj) )
		self.p1 = p1_obj
		p2_p = gdb.parse_and_eval("((HTwoParsers *) parser->env)->p2")
		p2_obj = top_level_parse.add_or_get_parser(p2_p)
		if p2_obj.name is None:
			p2_obj.name_parser( self.name_from_vtable(p2_obj) )
		self.p2 = p2_obj

	def __str__(self):
		return str([str(self.p1), str(self.p2)])

class EndEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return "[]"

class EndiannessEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_parser_p = gdb.parse_and_eval("((HParseEndianness *) parser->env)->p")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj
		endianness_val = gdb.parse_and_eval("((HParseEndianness *) parser->env)->endianness")
		self.endianness = int(endianness_val)

	def __str__(self):
		return str([str(self.member_parser), str(self.endianness)])

class EpsilonEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return "[]"

class IgnoreEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		ignored_p = gdb.parse_and_eval("(HParser*) parser->env")

		ignore_obj = top_level_parse.add_or_get_parser(ignored_p)
		if ignore_obj.name is None:
			ignore_obj.name_parser( self.name_from_vtable(ignore_obj) )
		self.member_parser = ignore_obj

	def __str__(self):
		return str([str(self.member_parser)])

# TODO: numeric parameter passed to parse_and_eval
class IgnoreSeqEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		self.member_parsers = []

		h_ignoreseq_t = gdb.lookup_type("HIgnoreSeq")
		h_ignoreseq_p_t = h_ignoreseq_t.pointer()

		parser_obj = gdb.parse_and_eval("*parser")
		p_env = parser_obj['env'].cast(h_ignoreseq_p_t).dereference()
		ignoreseq_p = gdb.parse_and_eval("(HIgnoreSeq*) parser->env")
		#seq_len = gdb.parse_and_eval("((HIgnoreSeq*) parser->env)->len")
		seq_len = p_env['len']
		#which = gdb.parse_and_eval("((HIgnoreSeq*) parser->env)->which")
		which = p_env['which']
		self.which = int(which)
		parsers = p_env['parsers']
		super().__init__(parser, top_level_parse)

		for index in range(0, seq_len):
			parser_p = parsers[index]
			parser_obj = top_level_parse.add_or_get_parser(parser_p)
			# TODO: name parsers on add?
			# TODO: move updates to parsers to TopLevelParse?
			if parser_obj.name is None:
				parser_obj.name_parser( self.name_from_vtable(parser_obj) )
			self.member_parsers.append(parser_obj)

	def __str__(self):
		return str([str(self.which), str([str(p) for p in self.member_parsers])])

class IndirectEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		member_parser_p = gdb.parse_and_eval("((HIndirectEnv*) parser->env)->parser")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj
		touched = gdb.parse_and_eval("((HIndirectEnv*) parser->env)->touched")
		self.touched = bool(touched)
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return str([str(self.member_parser), str(self.touched)])

class IntRangeEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_parser_p = gdb.parse_and_eval("((HRange*) parser->env)->p")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj
		lower_val = gdb.parse_and_eval("((HRange*) parser->env)->lower")
		self.lower = int(lower_val)
		upper_val = gdb.parse_and_eval("((HRange*) parser->env)->upper")
		self.upper = int(upper_val)


	def __str__(self):
		return str([str(self.member_parser), str(self.lower), str(self.upper)])

class LengthValueEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		length_parser_p = gdb.parse_and_eval("((HLenVal *) parser->env)->length")
		length_parser_obj = top_level_parse.add_or_get_parser(length_parser_p)
		if length_parser_obj.name is None:
			length_parser_obj.name_parser( self.name_from_vtable(length_parser_obj) )
		self.length_parser = length_parser_obj

		value_parser_p = gdb.parse_and_eval("((HLenVal *) parser->env)->value")
		value_parser_obj = top_level_parse.add_or_get_parser(value_parser_p)
		if value_parser_obj.name is None:
			value_parser_obj.name_parser( self.name_from_vtable(value_parser_obj) )
		self.value_parser = value_parser_obj

	def __str__(self):
		return str([str(self.length_parser), str(self.value_parser)])

class ManyEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		h_repeat_p = gdb.parse_and_eval("(HSequence*) parser->env")
		self.count = gdb.parse_and_eval("((HRepeat *) parser->env)->count")
		# TODO: resolve these pointers, make parser objects
		self.separator = gdb.parse_and_eval("((HRepeat *) parser->env)->sep")
		self.p = gdb.parse_and_eval("((HRepeat *) parser->env)->p")
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return "P: " + str(self.p) + ", sep: " + str(self.separator) + ", count: " + str(self.count)

class NotEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_p = gdb.parse_and_eval("(HParser*) parser->env")
		parser_obj = top_level_parse.add_or_get_parser(member_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj

	def __str__(self):
		return str(self.member_parser)

class NothingEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return "[]"

class OptionalEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_p = gdb.parse_and_eval("(HParser*) parser->env")
		parser_obj = top_level_parse.add_or_get_parser(member_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj

	def __str__(self):
		return str(self.member_parser)

# TODO: numeric parameter passed to parse_and_eval
class PermutationEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		super().__init__(parser, top_level_parse) # TODO: maybe move self.parser to base class. otherwise, is this needed?
		self.parser = parser
		self.member_parsers = []

		#frame = gdb.selected_frame()
		h_sequence_t = gdb.lookup_type("HSequence")
		h_sequence_p_t = h_sequence_t.pointer()

		# Get the parser object
		parser_obj = gdb.parse_and_eval("*parser")
		# Get parser_obj.env, cast it to HSequence *, then dereference
		p_env = parser_obj['env'].cast(h_sequence_p_t).dereference()
		# Get length of HSequence
		num_parsers = p_env['len']
		# Get the array of parsers in the sequence
		p_array = p_env['p_array']

		for index in range(0, num_parsers):
			parser_p = p_array[index]
			parser_obj = top_level_parse.add_or_get_parser(parser_p)
			if parser_obj.name is None:
				parser_obj.name_parser( self.name_from_vtable(parser_obj) )
			self.member_parsers.append(parser_obj)

	def __str__(self):
		return str([str(parser) for parser in self.member_parsers])

class SeekEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		offset_val = gdb.parse_and_eval("((HSeek *) parser->env)->offset")
		self.offset = int(offset_val)
		whence_val = gdb.parse_and_eval("((HSeek *) parser->env)->whence")
		self.whence = int(whence_val)
		#TODO: enum for whence?
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return str([str(self.offset), str(self.whence)])

class SkipEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		skip_val = gdb.parse_and_eval("((uintptr_t) parser->env)")
		self.skip = int(skip_val)
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return str(self.skip)

class TellEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return "[]"

# TODO: numeric parameter passed to parse_and_eval
class SequenceEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		super().__init__(parser, top_level_parse)
		self.parser = parser
		self.member_parsers = []

		#frame = gdb.selected_frame()
		h_sequence_t = gdb.lookup_type("HSequence")
		h_sequence_p_t = h_sequence_t.pointer()

		# Get the parser object
		parser_obj = gdb.parse_and_eval("*parser")
		# Get parser_obj.env, cast it to HSequence *, then dereference
		p_env = parser_obj['env'].cast(h_sequence_p_t).dereference()
		# Get length of HSequence
		num_parsers = p_env['len']
		# Get the array of parsers in the sequence
		p_array = p_env['p_array']

		for index in range(0, num_parsers):
			parser_p = p_array[index]
			parser_obj = top_level_parse.add_or_get_parser(parser_p)
			if parser_obj.name is None:
				parser_obj.name_parser( self.name_from_vtable(parser_obj) )
			self.member_parsers.append(parser_obj)

	def __str__(self):
		return str([str(parser) for parser in self.member_parsers])

class TokenEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser

		str_p = gdb.parse_and_eval("((HToken *) parser->env)->str")
		self.token = str_p
		str_len = gdb.parse_and_eval("((HToken *) parser->env)->len")
		self.str_len = int(str_len)
		#TODO: get str_len bytes from memory and save it
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return str(hex(self.token))

class UnimplemenetedEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

	def __str__(self):
		return "[]"

# TODO: print key as string if applicable
class GetEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_parser_p = gdb.parse_and_eval("((HStoredValue *) parse->env)->p")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj
		key_p = gdb.parse_and_eval("((HStoredValue *) parse->env)->key")
		self.key = int(key_p)

	def __str__(self):
		return str([str(self.member_parser), str(hex(self.key))])

class PutEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_parser_p = gdb.parse_and_eval("((HStoredValue *) parse->env)->p")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj
		key_p = gdb.parse_and_eval("((HStoredValue *) parse->env)->key")
		self.key = int(key_p)

	def __str__(self):
		return str([str(self.member_parser), str(hex(self.key))])

class WhitespaceEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		member_parser_p = gdb.parse_and_eval("(HParser *) parse->env")
		parser_obj = top_level_parse.add_or_get_parser(member_parser_p)
		if parser_obj.name is None:
			parser_obj.name_parser( self.name_from_vtable(parser_obj) )
		self.member_parser = parser_obj

	def __str__(self):
		return str([str(self.member_parser)])

class XorEnv(HParserEnv):
	def __init__(self, parser, top_level_parse):
		self.parser = parser
		super().__init__(parser, top_level_parse)

		p1_p = gdb.parse_and_eval("((HTwoParsers *) parser->env)->p1")
		p1_obj = top_level_parse.add_or_get_parser(p1_p)
		if p1_obj.name is None:
			p1_obj.name_parser( self.name_from_vtable(p1_obj) )
		self.p1 = p1_obj
		p2_p = gdb.parse_and_eval("((HTwoParsers *) parser->env)->p2")
		p2_obj = top_level_parse.add_or_get_parser(p2_p)
		if p2_obj.name is None:
			p2_obj.name_parser( self.name_from_vtable(p2_obj) )
		self.p2 = p2_obj

	def __str__(self):
		return str([str(self.p1), str(self.p2)])


vtable_to_env = {
	'action_vt': ActionEnv,
	'and_vt': AndEnv,
	'attr_bool_vt': AttrBoolEnv,
	'bind_vt': BindEnv,
	'bits_vt': BitsEnv,
	'butnot_vt': ButNotEnv,
	'ch_vt': ChEnv,
	'charset_vt': CharsetEnv,
	'choice_vt': ChoiceEnv,
	'difference_vt': DifferenceEnv,
	'end_vt': EndEnv,
	'endianness_vt': EndiannessEnv,
	'epsilon_vt': EpsilonEnv,
	'ignore_vt': IgnoreEnv,
	'ignoreseq_vt': IgnoreSeqEnv,
	'indirect_vt': IndirectEnv,
	'int_range_vt': IntRangeEnv,
	'length_value_vt': LengthValueEnv,
	'many_vt': ManyEnv,
	'not_vt': NotEnv,
	'nothing_vt': NothingEnv,
	'optional_vt': OptionalEnv,
	'permutation_vt': PermutationEnv,
	'seek_vt': SeekEnv,
	'skip_vt': SkipEnv,
	'tell_vt': TellEnv,
	'sequence_vt': SequenceEnv,
	'token_vt': TokenEnv,
	'unimplemented_vt': UnimplemenetedEnv,
	'get_vt': GetEnv,
	'put_vt': PutEnv,
	'whitespace_vt': WhitespaceEnv,
	'xor_vt': XorEnv
	}

# TODO: runtime address -> env class map

# When given a Parser object, decompose_parser() deduces its type from the vtable, and returns the appropriate HParserEnv subclass, containing member parsers and args
# Not sure what to name it. Alternatives:
# ParserDisassembler?
# ParserExploder?
# ParserDecomposer?

class ParserDecombinator:
	def __init__(self, vt_types):
		# TODO: make a better data structure than vtable_to_env + parser_name_defaults + this
		self.vt_envs = { k : (v,vtable_to_env[k]) for k,v in parser_name_defaults.items() }
		self.vt_types = vt_types

	def decompose_parser(self, parser, top_level_parse):
		# Sadly, this is stringly typed for now
		parser_addr = parser.address
		vtable_p = gdb.parse_and_eval("((HParser*) " + str(parser_addr) + ")->vtable")
		try:
			vtable_sym = self.vt_types.lookup_by_address(vtable_p)
			envClass = vtable_to_env[vtable_sym.name]
		except KeyError:
			print("Unknown vtable: " + str(vtable_p))
			return None

		return envClass(parser, top_level_parse)
